import pytest
from madcubapy.operations.maps import measure_noise


def test_invalid_object_type_measure_noise():
    with pytest.raises(TypeError):
        measure_noise("string")
    with pytest.raises(TypeError):
        measure_noise(999)


import astropy
from astropy.table import Table
from datetime import datetime
import getpass
import numpy as np
import os
from pathlib import Path

__all__ = [
    'MadcubaFits',
]

class MadcubaFits:
    """
    A basic class describing a MADCUBA FITS object with its history file.

    The MadcubaFits class contains the only object shared by MADCUBA maps,
    cubes, and spectra: a table describing the history file exported
    by MADCUBA.

    Parameters
    ----------
    hist : `~astropy.table.Table`
        Table containing the history information of the FITS file, which is
        stored in a separate *_hist.csv* file.

    """

    def __init__(
        self,
        hist=None,
    ):
        if hist is not None and not isinstance(hist, astropy.table.Table):
            raise TypeError(
                "The hist attribute must be an astropy Table")
        self._hist = hist

    @property
    def hist(self):
        """
        `~astropy.table.Table` : Table containing the history information.
        """
        return self._hist

    @hist.setter
    def hist(self, value):
        if value is not None and not isinstance(value, astropy.table.Table):
            raise TypeError(
                "The hist attribute must be an astropy Table")
        self._hist = value

    def add_hist(self, filename):
        """
        Load the history table from a csv file.

        Parameters
        ----------
        filename : `~str` or `~pathlib.Path`
            Path of the history csv file.

        """
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"file not found.")
        self._hist = Table.read(filename, format='csv')

    def _update_hist(self, action, **hist_keys):
        """
        Add a new row with information to the hist table.

        Parameters
        ----------
        faction : `~str`
            String describing the performed action.

        Other Parameters
        ----------------
        **hist_keys : dict
            Additional kwargs passed to the new row of the hist table.
            Allowed arguments for columns are: FITS, Type, FROM_ROW, TO_ROW,
            and Roi_Range.

        """
        if not self._hist:
            raise ValueError(
                f"This object does not have a history table."
            )
        # Fixed keys
        index = self._hist[-1]['Index'] + 1
        current_datetime = datetime.now()
        user = getpass.getuser()
        formatted_datetime = (current_datetime.strftime(f"%Y-%m-%dT%H:%M:%S")
                            + f".{current_datetime.microsecond // 1000:03d}")
        # Define the default values for the new_row dictionary
        new_row = {
            'Index': index,
            'FITS': 'C',
            'Macro': f'//PYTHON: {action}',
            'Type': 'Py',
            'FROM_ROW': np.ma.masked,
            'TO_ROW': np.ma.masked,
            'Roi_Range': np.ma.masked,
            'User': user,
            'Date': formatted_datetime,
        }
        all_keys = list(new_row)
        provided_keys = list(hist_keys)
        protected_keys = ['Index', 'Macro', 'User', 'Date']
        # Check if any of the protected keys are in hist_keys and raise an error
        for key in provided_keys:
            if key in protected_keys:
                raise ValueError(
                    f"'{key}' is a protected key and cannot be set manually."
                )
        # Ensure only the allowed keys are in hist_keys
        allowed_keys = [item for item in all_keys if item not in protected_keys]
        invalid_keys = [item for item in provided_keys if item not in allowed_keys]
        if invalid_keys:
            raise ValueError((f"Invalid key(s) provided: {', '.join(invalid_keys)}. "
                            + f"Only {', '.join(allowed_keys)} are allowed."))
        # Update the new_row dictionary with any valid hist_keys
        new_row.update(hist_keys)
        # Add new row to history
        self._hist.add_row(new_row)

    def __repr__(self):
        # If hist is None, display that it's missing
        if self._hist is None:
            hist_repr = "None"
        # If hist is present, display a summary of the table
        else: hist_repr = (
            f"<Table length={len(self._hist)} rows, " +
            f"{len(self._hist.columns)} columns>"
        )
        return f"<MadcubaFits(hist={hist_repr})>"



from astropy.table import Table
import astropy.units as u
from astropy.utils.diff import report_diff_values
import numpy as np
import pytest
from madcubapy.io.spectracontainer import SpectraContainer
from madcubapy.io.spectracontainer import parse_row_spectral_axis
from madcubapy.utils.spectral import create_spectral_array

@pytest.fixture
def example_init_spec():
    bintable = Table({
        'DATA': [np.array([[1, 2],[3, 4]]), np.array([[5, 6],[7, 8]])],
        'RESTFRQ': [1, 2],
        'CHANNELS': [5, 5],
        'CDELT3': [3, 3],
        'CRPIX3': [3, 3],
        'CRVAL3': [3, 3],
        'BUNIT': ['Jy    ', 'Jy    '],
    })
    hist = Table({
        'user': ['dh', 'dh'],
        'command': ['copy()', 'paste()'],
        'date': ['20241121', '20241122']
    })
    return SpectraContainer(bintable, hist)

@pytest.fixture
def example_init_nohist_spec():
    bintable = Table({
        'DATA': [np.array([[1, 2],[3, 4]]), np.array([[5, 6],[7, 8]])],
        'RESTFRQ': [1, 2],
        'CHANNELS': [5, 5],
        'CDELT3': [3, 3],
        'CRPIX3': [3, 3],
        'CRVAL3': [3, 3],
        'BUNIT': ['Jy    ', 'Jy    '],
    })
    return SpectraContainer(bintable)

@pytest.fixture
def example_read_spec():
    # Create and return a Map instance to be used in tests
    return SpectraContainer.read(
        "madcubapy/io/tests/data/IRAS16293_position_8_TM2_spectra.spec"
    )

def test_read_spec(example_read_spec):
    assert isinstance(example_read_spec.bintable, Table)
    assert isinstance(example_read_spec.hist, Table)
    assert example_read_spec.filename is not None

def test_invalid_file():
    with pytest.raises(FileNotFoundError):
        SpectraContainer.read("nonexistent_file.spec")

def test_parse_row_spectral_axis(example_read_spec):
    bintable = parse_row_spectral_axis(example_read_spec.bintable[0])
    assert isinstance(bintable, np.ndarray) or isinstance(bintable, u.Quantity)

def test_spectral_container_generate_spectral_axes(example_init_spec):
    # bintable without units
    example_init_spec._generate_spectral_axes()
    assert (example_init_spec.bintable[0]['XAXIS'].all()
            == np.array([-3.,  0.,  3.,  6.,  9.]).all())
    # bintable with units
    example_init_spec.bintable['RESTFRQ'].unit = u.Hz
    example_init_spec._generate_spectral_axes()
    assert (example_init_spec.bintable[0]['XAXIS'].all()
            == np.array([-3.,  0.,  3.,  6.,  9.]).all() and
            example_init_spec.bintable['XAXIS'].unit == u.Hz)

def test_spectral_container_parse_data_units(example_init_spec):
    # correctly parsed units
    example_init_spec._parse_data_units()
    assert example_init_spec.bintable['DATA'].unit == u.Jy
    # unequal strings for BUNIT
    example_init_spec.bintable["BUNIT"] = ["Jy  ", "Jy"]
    example_init_spec._parse_data_units()
    assert example_init_spec.bintable['DATA'].unit == None

def test_copy_read(example_read_spec):
    spectra_container_copy = example_read_spec.copy()
    assert report_diff_values(
        spectra_container_copy.hist, example_read_spec.hist
    )
    assert report_diff_values(
        spectra_container_copy.bintable, example_read_spec.bintable
    )

def test_copy_init(example_init_spec):
    spectra_container_copy = example_init_spec.copy()
    assert report_diff_values(
        spectra_container_copy.hist, example_init_spec.hist
    )
    assert report_diff_values(
        spectra_container_copy.bintable, example_init_spec.bintable
    )

def test_copy_init_nohist(example_init_nohist_spec):
    spectra_container_copy = example_init_nohist_spec.copy()
    assert report_diff_values(
        spectra_container_copy.bintable, example_init_nohist_spec.bintable
    )
    assert spectra_container_copy.hist == example_init_nohist_spec.hist
    assert spectra_container_copy.hist == None



from madcubapy.visualization import get_input
import pytest

def test_invalid_object_type_get_input():
    with pytest.raises(TypeError):
        get_input("string")
    with pytest.raises(TypeError):
        get_input(999)


from astropy.nddata import CCDData
import matplotlib as mpl
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
import matplotlib.pyplot as plt
import numpy as np
import tkinter as tk

from madcubapy.io import MadcubaMap
from .wcsaxes_helpers import add_wcs_axes
from .wcsaxes_helpers import insert_colorbar

__all__ = [
    'get_input',
]


def get_input(obj, **kwargs):
    """
    Get the pixel coordinates of points selected by mouse clicks in a
    previously plotted figure or a map object.

    Parameters
    ----------
    obj : `~matplotlib.figure.Figure`, \
          `~astropy.nddata.CCDData`, or \
          `~madcubapy.io.MadcubaMap` object.
        Figure or Map object to show and from which to get coordinates.

    Returns
    -------
    selected_points : `numpy.array`
        Coordinates of the clicked points.

    Other Parameters
    ----------------    
    **kwargs
        Optional map visualization parameters passed to
        :func:`~madcubapy.visualization.add_wcs_axes` only if `obj` is a
        `~astropy.nddata.CCDData` or `~madcubapy.io.MadcubaMap` object.
        
    """
    
    if isinstance(obj, plt.Figure):
        return _get_input_from_figure(obj)
    elif isinstance(obj, CCDData) or isinstance(obj, MadcubaMap):
        return _get_input_from_map(obj, **kwargs)
    else:
        raise TypeError("Unsupported type. " +
            "Provide a Matplotlib Figure, MadcubaMap, or CCDData object.")


def _get_input_from_figure(fig):
    """
    Returns mouse click coordinates using a previously created figure.

    Parameters
    ----------
    fig : `~matplotlib.figure.Figure`
        Figure object to show and from which to get coordinates.

    Returns
    -------
    selected_points : `numpy.array`
        Coordinates of the clicked points.

    """
    
    original_backend = mpl.get_backend()

    try:
        # Change matplotlib backend to tkinter
        mpl.use("TkAgg", force=True)  # Switch to TkAgg
        # Bring figure back from the dead
        new_fig = plt.figure()
        new_manager = new_fig.canvas.manager
        new_manager.canvas.figure = fig
        fig.set_canvas(new_manager.canvas)
        # Get input from clicks
        selected_points = fig.ginput(
            n=0, timeout=0, mouse_add=1, mouse_pop=2, mouse_stop=3
        )
        # Close figure to prevent issues
        plt.close('all')
    except Exception as e:
        raise RuntimeError(
            f"Failed to execute operation with TkAgg backend: {e}"
        )
    finally:
        try:
            # Change matplotlib backend back to inline
            mpl.use(original_backend, force=True)  # Restore original backend
            plt.close("all")  # Close any figures to prevent issues
        except Exception as e:
            raise RuntimeError(
                f"Failed to restore original backend ({original_backend}): {e}"
            )

    return np.array(selected_points)


def _get_input_from_map(fitsmap, **kwargs):
    """
    Returns mouse click coordinates using a `~astropy.nddata.CCDData` or
    `~madcubapy.io.madcubaMap` object.

    Parameters
    ----------
    fitsmap : `~astropy.nddata.CCDData` or `~madcubapy.io.MadcubaMap`
        Map object to show and from which to get coordinates.

    Returns
    -------
    selected_points : `numpy.array`
        Coordinates of the clicked points.

    Other Parameters
    ----------------
    **kwargs
        Parameters to pass to :func:`~madcubapy.visualization.add_wcs_axes`.

    """

    # Create a Tkinter window
    root = tk.Tk()
    root.title("Sigma calculation")

    # Create a Matplotlib figure
    fig = mpl.figure.Figure(figsize=(7, 6), dpi=100)
    ax, img = add_wcs_axes(fig, 1, 1, 1, fitsmap=fitsmap, **kwargs)
    cbar = insert_colorbar(ax)

    # Prettier plot
    ax.set_title('Sigma calculation', fontsize=15, pad=20)
    ax.coords[0].set_axislabel("RA (ICRS)", fontsize=12)
    ax.coords[1].set_axislabel("DEC (ICRS)", fontsize=12)
    ax.coords[0].tick_params(which="major",
                            length=5)
    ax.coords[0].display_minor_ticks(True)
    ax.coords[0].set_ticklabel(size=11, exclude_overlapping=True)
    ax.coords[1].tick_params(which="major",
                            length=5)
    ax.coords[1].display_minor_ticks(True)
    ax.coords[1].set_ticklabel(size=11, exclude_overlapping=True,
                               rotation='vertical')
    cbar.ax.yaxis.label.set_fontsize(12)
    cbar.ax.tick_params(labelsize=11)

    # Lists to store selected points and plot markers
    selected_points = []
    plot_markers = []

    # Create a callback function for mouse clicks
    def onclick(event):
        nonlocal selected_points, plot_markers
        # Left-click to select points
        if event.button == 1:
            if event.xdata is not None and event.ydata is not None:
                # Add the clicked point to the current polygon
                selected_points.append((event.xdata, event.ydata))
                # Draw the clicked point and add it to the list
                marker = ax.plot(event.xdata, event.ydata,
                                 'red', marker='+', markersize=6)
                plot_markers.append(marker)
                # Refresh the canvas
                canvas.draw()
        # Middle-click to remove last selected point
        elif event.button == 2:
            if selected_points:
                # Remove last selected point
                selected_points.pop()
                # Remove last painted point from plot and list
                marker = plot_markers[-1][0]
                marker.remove()
                plot_markers.pop()
                # Refresh the canvas
                canvas.draw()
        # Right-click to finalize the current polygon
        elif event.button == 3:
            # Refresh the canvas
            _quit()

    # Function to clear all selected points
    def clear_points():
        nonlocal selected_points, plot_markers
        # Remove all selected points
        selected_points.clear()
        # Remove all painted points
        while plot_markers:
            marker = plot_markers[-1][0]
            marker.remove()
            plot_markers.pop()
        canvas.draw()

    # Exit function to close the window
    def _quit():
        root.quit()
        root.destroy()

    # Add close shortcut
    def onkeypress(event):
        if event.key == 'q':
            root.quit()
            root.destroy()

    # Create a Matplotlib canvas embedded within the Tkinter window
    canvas = FigureCanvasTkAgg(fig, master=root)
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    # Add matplotlib tk toolbar
    toolbar = NavigationToolbar2Tk(canvas, root)
    toolbar.update()
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    # Create a frame for buttons
    control_frame = tk.Frame(root)
    control_frame.pack(side=tk.BOTTOM, pady=5)
    # "Clear All" button
    clear_button = tk.Button(control_frame, text="Clear All", command=clear_points)
    clear_button.pack(side=tk.LEFT, padx=5)
    # "Finish" button
    finish_button = tk.Button(control_frame, text="Finish", command=_quit)
    finish_button.pack(side=tk.RIGHT, padx=5)

    # Connect the onclick and onkeypress events to the canvas
    canvas.mpl_connect('button_press_event', onclick)
    canvas.mpl_connect('key_press_event', onkeypress)
    
    # Start the Tkinter event loop
    root.mainloop()

    return np.array(selected_points)



# madcubapy/madcubaio/__init__.py

"""
The `madcubapy.io` package provides classes to read MADCUBA products in
python.
"""

from .madcubafits import *
from .madcubamap import *
from .spectracontainer import *


from astropy.nddata import CCDData
import astropy.units as u
from copy import copy
import matplotlib as mpl
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
import matplotlib.patches as patches
import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import numpy as np
import tkinter as tk

from madcubapy.io import MadcubaMap
from madcubapy.visualization import add_wcs_axes
from madcubapy.visualization import insert_colorbar

__all__ = [
    'measure_noise',
]


def measure_noise(
        fitsmap,
        statistic='std',
        **kwargs):
    """
    Measure the noise (sigma) in a map object by calculating the standard
    deviation (std) or root mean square (rms) inside several polygons selected
    by mouse clicks.

    - Left clicks create polygon vertices.
    - Right click closes the current polygon, and a subsequent left click
      starts a new polygon.

    Parameters
    ----------
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map object to analize.
    statistic : {'std', 'rms'}, optional
        Statistic to be used as sigma. Defaults to 'std' and can be changed at
        runtime via GUI buttons.

    Returns
    -------
    sigma : `float`
        Measured noise of the image in the same units as the data array.

    Other Parameters
    ----------------
    **kwargs
        Optional map visualization parameters passed to
        :func:`~madcubapy.visualization.add_wcs_axes`.

    """

    if not isinstance(fitsmap, CCDData) and not isinstance(fitsmap, MadcubaMap):
        raise TypeError(f"Unsupported type: {type(fitsmap)}. "
                        + "Provide a MadcubaMap or CCDData object.")

    if statistic != 'std' and statistic != 'rms':
        raise ValueError(f"Invalid input for statistic: '{statistic}'. "
                         + "Accepted values are 'std' or 'rms'.")

    if ('use_std' not in kwargs and
        'vmin' not in kwargs and
        'vmax' not in kwargs):
        kwargs['use_std'] = True

    # Create a Tkinter window
    root = tk.Tk()
    root.title("Measure sigma")

    # Create a Matplotlib figure
    # If I use mpl.pyplot.figure here and this function is used in a
    # notebook, an inline plot of the final figure will be shown.
    fig = mpl.figure.Figure(figsize=(7, 6), dpi=100)
    ax, img = add_wcs_axes(fig, 1, 1, 1, fitsmap=fitsmap, **kwargs)
    cbar = insert_colorbar(ax)

    # Prettier plot
    current_title = ax.get_title()
    ax.set_title(current_title, fontsize=13, pad=15)
    ax.coords[0].set_axislabel("RA (ICRS)", fontsize=12)
    ax.coords[1].set_axislabel("DEC (ICRS)", fontsize=12)
    ax.coords[0].tick_params(which="major",
                            length=5)
    ax.coords[0].display_minor_ticks(True)
    ax.coords[0].set_ticklabel(size=11, exclude_overlapping=True)
    ax.coords[1].tick_params(which="major",
                            length=5)
    ax.coords[1].display_minor_ticks(True)
    ax.coords[1].set_ticklabel(size=11, exclude_overlapping=True,
                               rotation='vertical')
    cbar.ax.yaxis.label.set_fontsize(12)
    cbar.ax.tick_params(labelsize=11)

    # Lists to store the polygons and arrays with pixel data
    polygon_paths = []
    current_polygon = []
    inside_pixels = []
    sigma = np.nan

    # Create a callback function for mouse clicks
    def onclick(event):
        nonlocal polygon_paths, current_polygon
        # Left-click to select points
        if event.button == 1:
            # Add the clicked point to the current polygon
            current_polygon.append((event.xdata, event.ydata))
            now_polygon = np.array(current_polygon)
            # Draw a small point at the first clicked point
            if len(now_polygon) == 1:
                ax.plot(event.xdata, event.ydata,
                        'white', marker='o', markersize=5)
            # Draw lines
            if len(now_polygon) > 1:
                ax.plot(now_polygon[-2:, 0], now_polygon[-2:, 1],
                        'white', lw=2, alpha=0.9)
                ax.plot(event.xdata, event.ydata,
                        'white', marker='o', markersize=5)
            # Refresh the canvas
            canvas.draw()
        # Right-click to finalize the current polygon
        elif event.button == 3 and current_polygon:
            polygon = np.array(current_polygon)
            # Draw the last side of the polygon
            closed_polygon = np.vstack((polygon, polygon[0]))
            ax.plot(closed_polygon[-2:, 0], closed_polygon[-2:, 1],
                    'white', lw=2, alpha=0.9)
            # Draw the polygon on the plot
            poly = patches.Polygon(xy=polygon, linewidth=2,
                                   edgecolor='white', facecolor='white',
                                   alpha=0.5)
            new_poly = copy(poly)
            ax.add_patch(new_poly)
            # Get the shape of the CCDData object
            if fitsmap.header['NAXIS'] == 2:
                height, width = fitsmap.data.shape
                fitsmap_data = fitsmap.data
            elif fitsmap.header['NAXIS'] == 3:
                freq, height, width = fitsmap.data.shape
                fitsmap_data = fitsmap.data[0,:,:]
            elif fitsmap.header['NAXIS'] == 4:
                freq, stokes, height, width = fitsmap.data.shape
                fitsmap_data = fitsmap.data[0,0,:,:]
            # Create a meshgrid of coordinates to create mask
            x, y = np.meshgrid(range(width), range(height))
            points = np.vstack((x.flatten(), y.flatten())).T
            mask = poly.contains_points(points, radius=0)
            mask = mask.reshape((height, width))
            # Calculate std
            new_data = copy(fitsmap_data)
            std = new_data[mask].std(ddof=1)
            # Paint std inside polygon
            x_text = polygon.T[0].min() + (polygon.T[0].max()
                                           - polygon.T[0].min()) / 2
            y_text = polygon.T[1].min() + (polygon.T[1].max() 
                                           - polygon.T[1].min()) / 2
            std_text = ax.text(x_text, y_text, f'{std:.2f}',
                               va='center', ha='center',
                               color='white', fontsize=13)
            std_text.set_path_effects(
                [PathEffects.withStroke(linewidth=1.5, foreground="0.5")]
            )
            # Add info on lists
            polygon_paths.append(polygon)
            inside_pixels.append(new_data[mask])
            # Reset the current polygon
            current_polygon = []
            # Refresh the canvas
            canvas.draw()

    # Close function
    def _quit():
        sigma = get_sigma()
        if not np.isnan(sigma):
            root.quit()
            root.destroy()
        else:
            root.quit()
            root.destroy()

    # Close shortcut
    def onkeypress(event):
        if event.key == 'q':
            _quit()

    # Clear all drawn and selected polygons
    def clear_polygons():
        nonlocal polygon_paths
        nonlocal current_polygon
        nonlocal inside_pixels
        nonlocal sigma
        # Reset sigma lists
        polygon_paths = []
        current_polygon = []
        inside_pixels = []
        sigma = np.nan
        # Remove sigma painted objects
        # Remove lines and scatter points
        for line in ax.lines:
            line.remove()
        # Remove patches
        for patch in ax.patches:
            patch.remove()
        # Remove texts
        for text in ax.texts:
            text.remove()
        canvas.draw()

    # Calculate sigma
    def get_sigma():
        nonlocal sigma
        all_data = np.array([])
        for inside_data in inside_pixels:
            all_data = np.append(all_data, inside_data)
        if statistic == 'std':
            sigma = np.std(all_data, ddof=1)
        elif statistic == 'rms':
            sigma = np.sqrt(np.mean(np.square(all_data)))
        return sigma

    # Change sigma statistic
    def change_to_std():
        nonlocal statistic
        statistic = 'std'
        sigma_button.config(relief=tk.SUNKEN)
        check_button.config(relief=tk.RAISED)
    def change_to_rms():
        nonlocal statistic
        statistic = 'rms'
        sigma_button.config(relief=tk.RAISED)
        check_button.config(relief=tk.SUNKEN)

    # Create a Matplotlib canvas embedded within the Tkinter window
    canvas = FigureCanvasTkAgg(fig, master=root)
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    # Show toolbar
    toolbar = NavigationToolbar2Tk(canvas, root)
    toolbar.update()
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    # Create a frame for the statistic buttons
    stat_frame = tk.Frame(root)
    stat_frame.pack(side=tk.TOP, pady=5)
    if statistic == 'std':
        sigma_button = tk.Button(stat_frame, text="STD",
                                relief='sunken', command=change_to_std)
        sigma_button.pack(side=tk.LEFT, padx=3)
        check_button = tk.Button(stat_frame, text="RMS",
                                command=change_to_rms)
        check_button.pack(side=tk.LEFT, padx=3)
    elif statistic == 'rms':
        sigma_button = tk.Button(stat_frame, text="STD",
                                command=change_to_std)
        sigma_button.pack(side=tk.LEFT, padx=3)
        check_button = tk.Button(stat_frame, text="RMS",
                                relief='sunken', command=change_to_rms)
        check_button.pack(side=tk.LEFT, padx=3)

    # Create a frame for main buttons
    sigma_frame = tk.Frame(root)
    sigma_frame.pack(side=tk.BOTTOM)
    clear_button = tk.Button(sigma_frame, text="Clear", command=clear_polygons)
    clear_button.pack(side=tk.LEFT, padx=3)
    abort_button_sigma = tk.Button(sigma_frame, text="Finish", command=_quit)
    abort_button_sigma.pack(side=tk.RIGHT, padx=3)

    # Connect the onclick and onkeypress events to the canvas
    canvas.mpl_connect('button_press_event', onclick)
    canvas.mpl_connect('key_press_event', onkeypress)
    
    # Start the Tkinter event loop
    tk.mainloop()

    return sigma



from astropy.nddata import CCDData
import astropy.stats as stats
import astropy.units as u
import matplotlib as mpl
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import tkinter as tk
import sys

from madcubapy.io import MadcubaFits
from madcubapy.io import MadcubaMap
from .wcsaxes_helpers import add_wcs_axes
from .wcsaxes_helpers import insert_colorbar

__all__ = [
    'quick_show',
    'are_equal',
]

def quick_show(
        fitsmap,
        **kwargs):
    """
    Show a map on a separate window.

    Parameters
    ----------
    fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map to be displayed.

    Other Parameters
    ----------------
    **kwargs
        Parameters to pass to :func:`matplotlib.pyplot.imshow`.

    """

    # Create tkinter window
    root = tk.Tk()
    root.wm_title("Quick map plot")

    # Create a Matplotlib figure
    fig = mpl.figure.Figure(figsize=(6,5), dpi=100)
    ax, img = add_wcs_axes(fig, 1, 1, 1, fitsmap=fitsmap, **kwargs)
    cbar = insert_colorbar(ax)
    
    # Display minor ticks
    ax.coords[0].display_minor_ticks(True)
    ax.coords[1].display_minor_ticks(True)

    # Exit function to close the window
    def _quit():
        root.quit()
        root.destroy()

    # Add close shortcut
    def onkeypress(event):
        if event.key == 'q':
            root.quit()
            root.destroy()
    
    # Create a Matplotlib canvas embedded within the Tkinter window
    canvas = FigureCanvasTkAgg(fig, master=root)
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    # Add matplotlib tk toolbar
    toolbar = NavigationToolbar2Tk(canvas, root)
    toolbar.update()
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

    # Create a frame for buttons
    control_frame = tk.Frame(root)
    control_frame.pack(side=tk.BOTTOM, pady=5)
    #Exit button
    clear_button = tk.Button(control_frame, text="Quit", command=_quit)
    clear_button.pack(side=tk.LEFT)

    # Connect the onclick and onkeypress events to the canvas
    canvas.mpl_connect('key_press_event', onkeypress)

    # Start the Tkinter event loop
    root.mainloop()


def are_equal(
        fitsmap_1,
        fitsmap_2,
        show_maps=False):
    """
    Compare two maps to check if they are equal.

    Parameters
    ----------
    fitsmap_1 : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map object to compare.
    fitsmap_2 : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Map object to compare.
    show_maps : `~bool`, default: False
        If true both maps are plotted alongside the residuals of their
        substraction.

    Returns
    -------
    `~bool`
        Returns True if the two maps are the same map.

    """
    
    print('Checking if the two maps are equal\n')
    # Check fitsmap types
    if not (isinstance(fitsmap_1, (MadcubaMap, CCDData)) and
            isinstance(fitsmap_2, (MadcubaMap, CCDData))):
        raise TypeError(f"Input maps must be {MadcubaMap} or {CCDData}")
    # Copy fitsmaps to avoid changing input objects attributes
    map_1 = fitsmap_1.copy()
    map_2 = fitsmap_2.copy()
    # Slice extra dimensions from data
    if map_1.header['NAXIS'] == 3:
        data_1 = map_1.data[0, :, :]
    elif map_1.header['NAXIS'] == 4:
        data_1 = map_1.data[0, 0, :, :]
    else:
        data_1 = map_1.data
    if map_2.header['NAXIS'] == 3:
        data_2 = map_2.data[0, :, :]
    elif map_2.header['NAXIS'] == 4:
        data_2 = map_2.data[0, 0, :, :]
    else:
        data_2 = map_2.data
    # Shape check to avoid errors trying to plot residuals
    if data_1.shape != data_2.shape:
        raise ValueError("Cannot compare images with different resolutions")
    # Convert map 2 to units of map 1 if possible
    if isinstance(map_2, MadcubaMap):
        try:
            map_2.convert_unit_to(map_1.unit)
        except u.UnitConversionError:
            raise u.UnitConversionError(
                (
                    f"Cannot compare images with unconvertible units: "
                    + f"'{map_1.unit}' and '{map_2.unit}'"
                )
            )
    else:
        try:
            map_2 = map_2.convert_unit_to(map_1.unit)
        except u.UnitConversionError:
            raise u.UnitConversionError(
                (
                    f"Cannot compare images with unconvertible units: "
                    + f"'{map_1.unit}' and '{map_2.unit}'"
                )
            )
    # Show maps
    if show_maps:
        # Create figure
        fig = plt.figure(figsize=(15,6))
        # Map 1
        ax1, img1 = add_wcs_axes(fig, 1, 3, 1, fitsmap=map_1, use_std=True)
        cbar1 = insert_colorbar(ax1, 'top')
        ax1.set_title('MAP 1', pad=60)
        ax1.coords[1].set_ticklabel(
            rotation=90,
            rotation_mode='default',
            va='center'
        )
        # Map 2
        ax2, img2 = add_wcs_axes(fig, 1, 3, 2, fitsmap=map_2, use_std=True)
        cbar2 = insert_colorbar(ax2, 'top')
        ax2.set_title('MAP 2', pad=60)
        ax2.coords[1].set_ticklabel(
            rotation=90,
            rotation_mode='default',
            va='center'
        )
        # Diff map
        data = data_1 - data_2
        wcs = map_1.wcs.celestial
        mean, median, std = stats.sigma_clipped_stats(data, sigma=3.0)
        vmin = median - 3 * std
        vmax = median + 3 * std
        norm = 'linear'
        # Figure calling
        ax3 = plt.subplot(1, 3, 3, projection=wcs)
        img3 = ax3.imshow(data, vmin=vmin, vmax=vmax, cmap='viridis',
                        origin='lower', norm=norm)
        ax3.coords[1].set_ticklabel(
            rotation=90,
            rotation_mode='default',
            va='center'
        )
        ax3.set_xlabel('RA (ICRS)')
        ax3.set_ylabel('DEC (ICRS)')
        cbar3 = insert_colorbar(ax3, 'top')
        ax3.set_title('Residuals (1-2)', pad=60)
    # Check
    if np.array_equal(data_1, data_2, equal_nan=True):
        return True
    else:
        return False



from setuptools import setup, find_packages

VERSION = '0.2.0' 
DESCRIPTION = 'The MADCUBA python package.'

# Setting up
setup(
    name="madcubapy", 
    version=VERSION,
    author="David Haasler García",
    author_email="dhaasler@cab.inta-csic.es",
    description=DESCRIPTION,
    long_description=open('PYPIREADME.rst', 'r').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/dhaasler/madcubapy',
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=[
        "astropy>=6.0.0",
        "matplotlib>=3.6.0",
        "numpy>=1.23.0",
        "scipy>=1.9.2"
    ],
    keywords=[
        'madcuba',
        'radio astronomy',
    ],
    classifiers= [
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)



import numpy as np
import pytest
from madcubapy.io.madcubamap import MadcubaMap
from madcubapy.operations.maps import stack_maps

@pytest.fixture
def example_madcuba_map():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "madcubapy/operations/maps/tests/data/IRAS16293_SO_2-1_moment0_madcuba.fits"
    )

def test_stack_maps(example_madcuba_map):
    sum_map = stack_maps(example_madcuba_map, example_madcuba_map)
    assert np.array_equal(sum_map.data, example_madcuba_map.data*2, equal_nan=True)
    assert (sum_map.hist[-1]["Macro"] ==
        "//PYTHON: Stack maps. Files: 'IRAS16293_SO_2-1_moment0_madcuba.fits', 'IRAS16293_SO_2-1_moment0_madcuba.fits'"
    )



import astropy
from astropy.nddata import CCDData
from madcubapy.io import MadcubaMap
from madcubapy.visualization import are_equal
from pathlib import Path
import pytest

@pytest.fixture
def example_madcuba_map():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "madcubapy/io/tests/data/IRAS16293_SO_2-1_moment0_madcuba.fits"
    )

def test_are_equal(example_madcuba_map):
    assert are_equal(example_madcuba_map, example_madcuba_map.ccddata)



from astropy.nddata import CCDData
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np

from madcubapy.io import MadcubaMap

__all__ = [
    'stack_maps',
]


def stack_maps(*fitsmaps):
    """
    Adds the data of multiple map objects (`~madcubapy.io.MadcubaMap` or
    `~astropy.nddata.CCDData`) together and returns the result as a new map
    object with the metadata of the first input map.
    
    Parameters
    ----------
    *fitsmaps : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        Any number of map objects to be added together.
    
    Returns
    -------
    combined_fitsmap : `~madcubapy.io.MadcubaMap` or `~astropy.nddata.CCDData`
        A new map object containing the sum of the data of the input maps.
    
    """

    # Ensure the input is not empty
    if not fitsmaps:
        raise ValueError("At least one map object must be provided.")
    # Check map object type
    first_type = type(fitsmaps[0])
    if not all(type(obj) == first_type for obj in fitsmaps):
        raise TypeError("Input parameters must be of the same type.")
    elif (type(fitsmaps[0]) != MadcubaMap and type(fitsmaps[0]) != CCDData):
        raise TypeError(
            f"Input parameters must be {MadcubaMap} or {CCDData} types."
        )
    # Ensure all CCDData objects have the same shape
    shape = fitsmaps[0].data.shape
    for fitsmap in fitsmaps[1:]:
        if fitsmap.data.shape != shape:
            raise ValueError("All maps must have the same shape.")
    # Ensure all CCDData objects have the same unit
    unit = fitsmaps[0].unit
    for fitsmap in fitsmaps[1:]:
        if fitsmap.unit != unit:
            fitsmap = fitsmap.to(unit)
    # Create and return the new map object with combined data and the metadata
    # from the first map object
    combined_data = sum(fitsmap.data for fitsmap in fitsmaps)
    if type(fitsmaps[0]) == MadcubaMap:
        # Create new CCDData if present
        if fitsmaps[0].ccddata:
            new_ccddata = CCDData(
                data=combined_data,
                unit=unit,
                meta=fitsmaps[0].ccddata.meta,
                wcs=fitsmaps[0].ccddata.wcs,
            )
        else:
            new_ccddata = None
        # Use history table from first object if present
        if fitsmaps[0].hist:
            new_hist = fitsmaps[0].hist.copy()
        else:
            new_hist = None
        # Create new madcubaMap
        new_madcubamap = MadcubaMap(
            data=combined_data,
            header=fitsmaps[0].header,
            wcs=fitsmaps[0].wcs,
            unit=unit,
            hist=new_hist,
            ccddata=new_ccddata,
            filename=fitsmaps[0].filename,
        )
        # Create history action
        if all(obj.filename is not None for obj in fitsmaps):
            stacked_maps_names = [obj.filename for obj in fitsmaps]
            update_action = ("Stack maps. Files: '" +
                             "', '".join(stacked_maps_names) + "'")
        else:
            update_action = (
                "Stack maps. Manually created objects with no files"
            )
        # Update history if present
        if new_madcubamap.hist:
            new_madcubamap._update_hist(update_action)
        return new_madcubamap
    else:
        return CCDData(combined_data, unit=unit,
                       meta=fitsmaps[0].meta, wcs=fitsmaps[0].wcs)

    return combined_fitsmap



from astropy.table import Table
import numpy as np
import pytest
from madcubapy.io.madcubafits import MadcubaFits

@pytest.fixture
def example_madcuba_fits_hist():
    # Create and return a Fits instance with hist table to be used in tests
    data = {
        'Index': [1, 2, 3],
        'FITS': ['C', 'C', 'C'],
        'Macro': ['Line 1', 'Line 2', 'Line 3'],
        'Type': ['Line 1', 'Line 2', 'Line 3'],
        'FROM_ROW': [np.ma.masked, np.ma.masked, np.ma.masked],
        'TO_ROW': [np.ma.masked, np.ma.masked, np.ma.masked],
        'Roi_Range': [np.ma.masked, np.ma.masked, np.ma.masked],
        'User': ['Line 1', 'Line 2', 'Line 3'],
        'Date': ['Line 1', 'Line 2', 'Line 3'],
    }
    table = Table(data)
    return MadcubaFits(hist=table)

def test_initialization_with_none():
    madcuba_fits = MadcubaFits()
    assert madcuba_fits.hist is None

def test_initialization_with_table():
    data = {'col1': [1, 2, 3], 'col2': [4, 5, 6]}
    table = Table(data)
    madcuba_fits = MadcubaFits(hist=table)
    assert isinstance(madcuba_fits.hist, Table)
    assert len(madcuba_fits.hist) == 3  # Check table length

def test_invalid_initialization_type():
    with pytest.raises(TypeError):
        MadcubaFits(hist="string")
    with pytest.raises(TypeError):
        MadcubaFits(hist=3)
    with pytest.raises(TypeError):
        MadcubaFits(hist=0.4)
    with pytest.raises(TypeError):
        MadcubaFits(hist=np.zeros(4,5))

def test_hist_setter():
    data = {'col1': [1, 2], 'col2': [3, 4]}
    table = Table(data)
    madcuba_fits = MadcubaFits()
    # Set hist to a valid table
    madcuba_fits.hist = table
    assert isinstance(madcuba_fits.hist, Table)
    assert len(madcuba_fits.hist) == 2
    # Set hist to None
    madcuba_fits.hist = None
    assert madcuba_fits.hist is None

def test_add_invalid_hist_file():
    madcuba_fits = MadcubaFits()
    with pytest.raises(FileNotFoundError):
        madcuba_fits.add_hist("nonexistent_file.csv")

def test_update_hist(example_madcuba_fits_hist):
    example_madcuba_fits_hist._update_hist(action="Fix temp", FITS='S')
    assert example_madcuba_fits_hist.hist[-1]['Index'] == 4
    assert example_madcuba_fits_hist.hist[-1]['FITS'] == 'S'
    assert example_madcuba_fits_hist.hist[-1]['Macro'] == '//PYTHON: Fix temp'
    assert example_madcuba_fits_hist.hist[-1]['Type'] == 'Py'
    assert example_madcuba_fits_hist.hist[-1]['FROM_ROW'] is np.ma.masked
    assert example_madcuba_fits_hist.hist[-1]['TO_ROW'] is np.ma.masked
    assert example_madcuba_fits_hist.hist[-1]['Roi_Range'] is np.ma.masked

def test_update_hist_protected(example_madcuba_fits_hist):
    with pytest.raises(ValueError):
        example_madcuba_fits_hist._update_hist(action="Fix temp", Index='S')
    with pytest.raises(ValueError):
        example_madcuba_fits_hist._update_hist(action="Fix temp", Macro='S')
    with pytest.raises(ValueError):
        example_madcuba_fits_hist._update_hist(action="Fix temp", User='S')
    with pytest.raises(ValueError):
        example_madcuba_fits_hist._update_hist(action="Fix temp", Date='S')

def test_update_hist_invalid(example_madcuba_fits_hist):
    with pytest.raises(ValueError):
        example_madcuba_fits_hist._update_hist(action="Fix temp", Any_name='S')

def test_update_hist_none():
    madcuba_fits = MadcubaFits()
    with pytest.raises(ValueError):
        madcuba_fits._update_hist(action="Fix temp")



import astropy
from astropy.table import Table
import astropy.units as u
import numpy as np
import os
from pathlib import Path
import zipfile

from .madcubafits import MadcubaFits
from madcubapy.utils.spectral import create_spectral_array

__all__ = [
    'SpectraContainer',
    'parse_row_spectral_axis',
]

class SpectraContainer(MadcubaFits):
    """
    A container for MADCUBA spectra, using the
    `~madcubapy.io.madcubafits.MadcubaFits` interface.

    This class is basically a wrapper to read MADCUBA exported spectra and
    their history files with astropy.

    Parameters
    ----------
    bintable : `~astropy.table.Table`
        Table containing the data of every spectra inside the *.spec* file
        alongside the info of their headers.
    hist : `~astropy.table.Table`
        Table containing the history information of the FITS file, which is
        stored in a separate *_hist.csv* file.
    filename : `~str`
        Name of the *.spec* file.

    Methods
    -------
    add_hist(*args)
        Load the history table from a csv file.

    """

    def __init__(
        self,
        bintable=None,
        hist=None,
        filename=None,
    ):
        # inherit hist
        super().__init__(hist)

        if bintable is not None and not isinstance(bintable, Table):
            raise TypeError(
                "The bintable must be an astropy Table")
        self._bintable = bintable

        if filename is not None and not isinstance(filename, str):
            raise TypeError("The filename must be a string.")
        self._filename = filename

    @property
    def bintable(self):
        """
        `~astropy.table.Table` : Table containing the data of every spectra
        inside the *.spec* file.
        """
        return self._bintable

    @bintable.setter
    def bintable(self, value):
        if value is not None and not isinstance(value, Table):
            raise TypeError(
                "The bintable must be an astropy Table")
        self._bintable = value

    @property
    def filename(self):
        """
        `~str` : Name of the *.spec* file.
        """
        return self._filename

    @filename.setter
    def filename(self, value):
        if value is not None and not isinstance(value, str):
            raise TypeError("The filename must be a string.")
        self._filename = value

    @classmethod
    def read(cls, filepath):
        """
        ``Classmethod`` to generate a
        `~madcubapy.io.spectracontainer.SpectraContainer` from a *.spec* file.

        Parameters
        ----------
        filepath : `~str`
            Name of spec file.

        """
        spec_filepath = filepath
        # Check if the spec file exists
        if not os.path.isfile(spec_filepath):
            raise FileNotFoundError(f"File {spec_filepath} not found.")
        # Open the ZIP file
        with zipfile.ZipFile(spec_filepath, "r") as zip_file:
            fits_files = [
                name for name in zip_file.namelist() if name.endswith(".fits")
            ]
            with zip_file.open(fits_files[0]) as internal_fits_file:
                bintable = Table.read(internal_fits_file, hdu=1)
            hist_files = [
                name for name in zip_file.namelist() if name.endswith("_hist.csv")
            ]
            with zip_file.open(hist_files[0]) as internal_hist_file:
                hist = Table.read(internal_hist_file, format='csv')
        filename_terms = str(filepath).split('/')
        filename = filename_terms[-1]
        # Return an instance of MadcubaFits
        spectra_container = cls(
            bintable=bintable,
            hist=hist,
            filename=filename,
        )
        spectra_container._generate_spectral_axes()
        spectra_container._parse_data_units()
        return spectra_container

    def copy(self):
        """
        Create a copy of the `~madcubapy.io.SpectraContainer`.
        """
        if self._bintable:
            new_bintable = self._bintable.copy()
        else:
            new_bintable = None
        if self._hist:
            new_hist = self._hist.copy()
        else:
            new_hist = None
        return SpectraContainer(
            bintable=new_bintable,
            hist=new_hist,
        )

    def _generate_spectral_axes(self):
        """
        Generate arrays for the spectral axes of every spectra inside the
        `~madcubapy.io.SpectraContainer` and add them as a new table column.
        """
        data = []
        for i in range(len(self.bintable)):
            spectral_axis = parse_row_spectral_axis(self.bintable[i])
            if isinstance(data, u.Quantity):
                data.append(spectral_axis.value)
            else: data.append(spectral_axis)
        self.bintable['XAXIS'] = data
        self.bintable['XAXIS'].unit = self.bintable['RESTFRQ'].unit
        
    def _parse_data_units(self):
        """
        Parse the BUNIT column values. If all are equal, set it as the unit for
        the DATA column.
        """
        if np.all(self.bintable["BUNIT"] == self.bintable["BUNIT"][0]):
            unit_code = self.bintable["BUNIT"][0]
            try: 
                unit = astropy.units.Unit(unit_code)
            except ValueError:
                print("Unit string for data could not be parsed")
                unit = unit_code
            finally:
                self.bintable["DATA"].unit = unit
        else: self.bintable["DATA"].unit = None
        
    def __repr__(self):
        # If hist is None, display that it's missing
        if self._hist is None:
            hist_repr = "hist=None"
        # If hist is present, display a summary of the table
        else:
            hist_repr = (
                f"hist=<Table length={len(self._hist)} rows, "
                + f"{len(self._hist.columns)} columns>"
            )
        if self._bintable is None:
            bintable_repr = "bintable=None"
        # If hist is present, display a summary of the table
        else:
            bintable_repr = (
                f"bintable=<Table length={len(self._hist)} rows, "
                + f"{len(self._hist.columns)} columns>"
            )

        return f"<SpectraContainer({bintable_repr}, {hist_repr})>"


def parse_row_spectral_axis(table_row):
    """
    Generate an array for the spectral axis of a spectrum inside a
    `~madcubapy.io.SpectraContainer`.

    Parameters
    ----------
    table_row : `astropy.table.Row`
        Row of a `~madcubapy.io.SpectraContainer`'s bintable. This is the data
        for a spectrum inside a MADCUBA FITS file.
    
    Returns
    -------
    spectral_array : `~numpy.ndarray` or `~astropy.units.Quantity`
        Returned spectral axis array with units if correctly parsed from the
        FITS file.

    """
    
    # Get spectrum data
    nchan = table_row['CHANNELS']
    cdelt3 = table_row['CDELT3']
    crval3 = table_row['CRVAL3']
    crpix3 = table_row['CRPIX3'] - 1  # fits to numpy
    if isinstance(table_row.table['RESTFRQ'].unit, astropy.units.UnitBase):
        spectral_unit = table_row.table['RESTFRQ'].unit
    else:
        spectral_unit = None
    spectral_array = create_spectral_array(nchan,cdelt3,crpix3,crval3)
    if spectral_unit:
        spectral_array = spectral_array * spectral_unit

    return spectral_array



import astropy
from astropy.nddata import CCDData
from madcubapy.io import MadcubaMap
from madcubapy.visualization import add_wcs_axes
from madcubapy.visualization import add_manual_wcs_axes
from madcubapy.visualization import add_colorbar
from madcubapy.visualization import insert_colorbar
from madcubapy.visualization import parse_clabel
import matplotlib as mpl
import matplotlib.pyplot as plt
from pathlib import Path
import pytest

@pytest.fixture
def example_madcuba_map():
    # Create and return a Map instance to be used in tests
    return MadcubaMap.read(
        "madcubapy/io/tests/data/IRAS16293_SO_2-1_moment0_madcuba.fits"
    )

def test_add_wcs_axes(example_madcuba_map):
    fig = plt.figure(figsize=(10,5))
    ax, img = add_wcs_axes(fig, 1, 2, 1, fitsmap=example_madcuba_map)
    assert isinstance(ax, astropy.visualization.wcsaxes.WCSAxes)
    assert isinstance(img, mpl.image.AxesImage)

def test_add_wcs_axes_without_arguments(example_madcuba_map):
    fig = plt.figure(figsize=(10,5))
    with pytest.raises(TypeError):
        add_wcs_axes(nrows=1, ncols=2, number=1, fitsmap=example_madcuba_map)
    with pytest.raises(TypeError):
        add_wcs_axes(fig=fig, nrows=1, ncols=2, number=1)

def test_add_wcs_axes_with_incorrect_types(example_madcuba_map):
    fig = plt.figure(figsize=(10,5))
    with pytest.raises(TypeError):
        add_wcs_axes(fig=3, nrows=1, ncols=2, number=1, fitsmap=example_madcuba_map)
    with pytest.raises(TypeError):
        add_wcs_axes(fig=fig, nrows=1, ncols=2, number=1, fitsmap=4)

def test_add_colorbar(example_madcuba_map):
    fig = plt.figure(figsize=(10,5))
    ax, img = add_wcs_axes(fig, 1, 2, 1, fitsmap=example_madcuba_map)
    cbar = add_colorbar(ax=ax, location='top', label='custom units')
    assert isinstance(cbar, mpl.colorbar.Colorbar)

def test_insert_colorbar(example_madcuba_map):
    fig = plt.figure(figsize=(10,5))
    ax, img = add_wcs_axes(fig, 1, 2, 1, fitsmap=example_madcuba_map)
    cbar = insert_colorbar(ax=ax, location='top', label='custom units')
    assert isinstance(cbar, mpl.colorbar.Colorbar)

def test_parse_clabel(example_madcuba_map):
    label = parse_clabel(example_madcuba_map)
    assert label == r'$I \ {\rm (Jy \ beam^{-1} \ m \ s^{-1})}$'
